WEAKNESS_BEFORE_STOP_WORDS = r' ((has|contain[s]?|introduc(es|ed|ing)) (an |a |)|(was|were) (found|discovered) to contain a )'
WEAKNESS_PATTERN = r'(?P<weakness>.*?)'
WEAKNESS_AFTER_STOP_WORDS = r'( ((V|v)ulnerabilit(y|ies)|issue[s]?)( (was|were) (discovered|found)|)| exists|)'

COMPONENT_BEFORE_STOP_WORDS = r' (in (the |its |)|(starting |)at )'
COMPONENT_PATTERN = r'(?P<component>.+?)'
COMPONENT_AFTER_STOP_WORDS = r'(file|component|call|feature|function|method|process|implementation)[s]?'

QUOTES_PATTERN = r'[`\'"\u2018\u2019\u201C\u201D]'
DATE_VERSION_PATTERN = r'(before|through) \d{4}-\d{2}-\d{2}'
YEAR_PATTERN = r'(?:200\d|201\d|202\d)'
VERSION_PATTERN = r'\b(?:v\.|v|R)*(?:\d+(?:\.(?:\d+|x))+)(?:\-\d|a)*\b'
COMMIT_VERSION_PATTERN = r'\b[0-9a-f]{5,40}\b'

CONFIGURATION_BEFORE_STOP_WORDS = r' (in|of|on) (some |the |a |)'
CONFIGURATION_PATTERN = (r'(?P<configuration>[a-zA-Z0-9-\/\(\),]{2,}.+?(?:' + VERSION_PATTERN + r'|' + COMMIT_VERSION_PATTERN
                         + r'|' + DATE_VERSION_PATTERN + r'|' + YEAR_PATTERN + r')(?: and (?:before|prior|(possibly |)earlier|below|up to and including)| plugin|.+?version[s]?|))')
CONFIGURATION_AFTER_STOP_WORDS = r'( for(?: [a-zA-Z0-9\!\-]+){1,3}|)'

ATTACKER_BEFORE_STOP_WORDS = r'(,|) (could |which (could |)|may |might |that |)(allow|enabl[e]?|(exploitable|used) by)(s|ed|ing|) (a |an |any |)'
ATTACKER_PATTERN = r'(?P<attacker>.*?(attacker|user|server|administrator|admin|subscriber|adversary)[s]?)'
ATTACKER_AFTER_STOP_WORDS = r'(,|)( with.+?(permission|privilege|role|access|right)[s]?( and above| as low.+?|)| (((who are |)able|with the ability) to).+?| such as.+?| permitted|)'

ATTACK_BEFORE_STOP_WORDS = r' (to (perform|execute) )'
ATTACK_PATTERN = r'(?P<attack>.*?(attack)[s]?)'

CONDITION_BEFORE_STOP_WORDS = r'(,|) (if |where |(even |)when )'
CONDITION_PATTERN = r'(?P<condition>.+?)'

IMPACT_ACTIONS = [
    "cause", "bypass", "gain", "hijack", "escalate", "inject", "trigger", "upload", "update", "connect", "send",
    "redirect", "replace", "change", "create", "read", "enumerate", "impersonate", "call", "request", "delete", "steal",
    "define", "lead", "check", "access", "execute", "run", "force", "crash", "overwrite", "corrupt", "write", "take",
    "make", "conduct", "perform", "obtain", "achieve", "remotely", "arbitrarily", "trick", "copy", "replace", "accept",
    "disable", "initiate", "rebuild", "include", "modify", "add", "edit", "configure", "have", "guess", "view", "list",
    "retrieve", "insert", "disclose", "reset", "craft", "open", "discover", "determine", "visit", "introduce",
    "harvest", "exfiltrate", "clear", "supply", "leak", "use"
]

IMPACT_ACTIONS_PATTERN = '|'.join(IMPACT_ACTIONS)

IMPACT_BEFORE_STOP_WORDS = r'(,|) ((to|which( could)) (' + IMPACT_ACTIONS_PATTERN + '|potentially( enable|))[s]?(,|) (a |)|resulting in |leading to |causing a |and can lead to )'
IMPACT_PATTERN = r'(?P<impact>.+?)'

VECTOR_BEFORE_STOP_WORDS = r'(,|) (via|(as demonstrated |)by) (an |a |vectors |)'
VECTOR_PATTERN = r'(?P<vector>.+?)'

ROOT_CAUSE_BEFORE_STOP_WORD = r'(,|) (does not|because( of| a| the|)|(due|fail) to|insecurely) '
ROOT_CAUSE_PATTERN = r'(?P<root_cause>.+?)'


FUNCTION_NAME_PATTERN = r'[a-zA-Z_][a-zA-Z0-9_:\(\)]+'
EXTENSION_PATTERN = r"\.[a-zA-Z0-9]{1,4}"
DIR_FILE_NAME_PATTERN = r"[a-zA-Z0-9_-]+"
FILE_PATTERN = DIR_FILE_NAME_PATTERN + EXTENSION_PATTERN
RELATIVE_PATH_PATTERN = DIR_FILE_NAME_PATTERN + r"(?:/(?:" + DIR_FILE_NAME_PATTERN + r"))+" + EXTENSION_PATTERN


REPLACEMENT_PHRASES = {
    r'via the (' + FUNCTION_NAME_PATTERN + r') function in the component (.+?)\.$': r'in the \1 function at \2 component.',
    r'via the (' + FUNCTION_NAME_PATTERN + r') (function|component) (at|in) (.+?)\.$': r'in the \1 at \4.',
    r'via the (function|component) (' + FUNCTION_NAME_PATTERN + r') (at|in) (.+?)\.$': r'in the \2 \1 at \4.',
    r'via (' + FUNCTION_NAME_PATTERN + r') class in (.+?)\.$': r'in the \1 at \2.',
    r'via the (' + FUNCTION_NAME_PATTERN + r') function.$': r'in the \1 function.',
    r'via the function (' + FUNCTION_NAME_PATTERN + r').$': r'in the \1 function.',
    r'via the component (.+?)\.$': r'in the \1.',
    r'in function (' + FUNCTION_NAME_PATTERN + r') in (.+?) in ': r'in \1 function at \2 in ',
    r'in function (' + FUNCTION_NAME_PATTERN + r') in ' + CONFIGURATION_PATTERN: r'in \1 function in \2',
    r'at (.+?) in (' + FUNCTION_NAME_PATTERN + r') in ' + CONFIGURATION_PATTERN: r'in \2 at \1 in \3',
    r'out of bounds': r'out-of-bounds',
    r'has been identified in': r'exists in',
    r'(that |)leads to( a|)': r'leading to',
    r'in the context of': r'within',
    r'that( can|) result[s]?( in|)': r'resulting in',
    r'^There is a': r'A',
    r' to to ': ' to ',
    r'in ' + QUOTES_PATTERN + r'([a-zA-Z0-9-\/\(\),]{2,})' + QUOTES_PATTERN: r'in \1',
    r' <= (' + VERSION_PATTERN + ')': r' up to and including \1',
    r'is vulnerable to': r'has',
    r' attackers to Injecting ': ' attackers to inject ',
    r'resulting in (.*?) vulnerability': r'introduces \1 vulnerability',
    r'(\d+), ([a-z]+)': r'\1 \2',
    r'^In ' + CONFIGURATION_PATTERN + r' there is a': r'\1 has a',
    r'^(?:In |)' + CONFIGURATION_PATTERN + r' ((?:a |).+?)(?: vulnerability|) exists in': r'\1 has \4 in',
    r'^(' + FUNCTION_NAME_PATTERN + r') (in) (' + RELATIVE_PATH_PATTERN + r')' + CONFIGURATION_BEFORE_STOP_WORDS : r'\1 \2 \3 file \4 ',
    r'(' + FUNCTION_NAME_PATTERN + r') function in (' + RELATIVE_PATH_PATTERN + r')' + CONFIGURATION_BEFORE_STOP_WORDS : r'\1 function at \2 file \3 ',
    r'(' + FUNCTION_NAME_PATTERN + r') function in (' + FILE_PATTERN + r')' + CONFIGURATION_BEFORE_STOP_WORDS : r'\1 function at \2 file \3 ',
    r'(' + FUNCTION_NAME_PATTERN + r') (in) (' + FILE_PATTERN + r')' + CONFIGURATION_BEFORE_STOP_WORDS : r'\1 at \3 file \4 ',
    r'^(' + RELATIVE_PATH_PATTERN + r')' + CONFIGURATION_BEFORE_STOP_WORDS: r'\1 file \2 ',
    r'^(' + FILE_PATTERN + r')' + CONFIGURATION_BEFORE_STOP_WORDS: r'\1 file \2 '
}

NOISE_PATTERNS = [
    r'\s+\(.*?\)',
    r'\s+\[CWE-\d{1,4}\]',
    r'( of type|)\s+CWE-\d{1,4}',
    r' a different vulnerability than CVE-\d{4}-\d{4,7}',
    r' a related issue to CVE-\d{4}-\d{4,7}',
    r' and CVE-\d{4}-\d{4,7}',
    r' fixed in ' + VERSION_PATTERN,
    r'can be abused '
]



from ...common import (CONFIGURATION_PATTERN, COMPONENT_PATTERN, CONFIGURATION_BEFORE_STOP_WORDS,
                       COMPONENT_AFTER_STOP_WORDS, ATTACKER_BEFORE_STOP_WORDS,
                       ATTACKER_PATTERN, ATTACKER_AFTER_STOP_WORDS, IMPACT_BEFORE_STOP_WORDS, IMPACT_PATTERN,
                       VECTOR_BEFORE_STOP_WORDS, VECTOR_PATTERN, ROOT_CAUSE_BEFORE_STOP_WORD, ROOT_CAUSE_PATTERN)


# component config root_cause
COMPONENT_CONFIG_CAUSE_PATTERN = (r'(?:The |)' + COMPONENT_PATTERN + COMPONENT_AFTER_STOP_WORDS
                                     + CONFIGURATION_BEFORE_STOP_WORDS + CONFIGURATION_PATTERN
                                     + ROOT_CAUSE_BEFORE_STOP_WORD + ROOT_CAUSE_PATTERN)

# component config root_cause attacker impact vector
DETAILS_6_1 = (COMPONENT_CONFIG_CAUSE_PATTERN + ATTACKER_BEFORE_STOP_WORDS + ATTACKER_PATTERN +
               ATTACKER_AFTER_STOP_WORDS + IMPACT_BEFORE_STOP_WORDS + IMPACT_PATTERN + VECTOR_BEFORE_STOP_WORDS +
               VECTOR_PATTERN + r'\.$')

# component config root_cause attacker impact
DETAILS_5_1 = (COMPONENT_CONFIG_CAUSE_PATTERN + ATTACKER_BEFORE_STOP_WORDS + ATTACKER_PATTERN +
               ATTACKER_AFTER_STOP_WORDS + IMPACT_BEFORE_STOP_WORDS + IMPACT_PATTERN + r'\.$')


COMPONENT_CONFIG_CAUSE_PATTERNS = {
    '6_1': DETAILS_6_1,
    '5_1': DETAILS_5_1,
}



from ...common import (CONFIGURATION_PATTERN, COMPONENT_PATTERN, CONFIGURATION_BEFORE_STOP_WORDS, CONDITION_PATTERN,
                       COMPONENT_AFTER_STOP_WORDS, ATTACKER_BEFORE_STOP_WORDS, CONDITION_BEFORE_STOP_WORDS,
                       ATTACKER_PATTERN, ATTACKER_AFTER_STOP_WORDS, IMPACT_BEFORE_STOP_WORDS, IMPACT_PATTERN,
                       VECTOR_BEFORE_STOP_WORDS, VECTOR_PATTERN, ROOT_CAUSE_BEFORE_STOP_WORD, ROOT_CAUSE_PATTERN)


# component config attacker
COMPONENT_CONFIG_ATTACKER_PATTERN = (r'(?:The |)' + COMPONENT_PATTERN + COMPONENT_AFTER_STOP_WORDS
                                     + CONFIGURATION_BEFORE_STOP_WORDS + CONFIGURATION_PATTERN
                                     + ATTACKER_BEFORE_STOP_WORDS + ATTACKER_PATTERN + ATTACKER_AFTER_STOP_WORDS)

# component config attacker impact vector root_cause
DETAILS_6_1 = (COMPONENT_CONFIG_ATTACKER_PATTERN + IMPACT_BEFORE_STOP_WORDS + IMPACT_PATTERN +
               VECTOR_BEFORE_STOP_WORDS + VECTOR_PATTERN + ROOT_CAUSE_BEFORE_STOP_WORD + ROOT_CAUSE_PATTERN + r'\.$')

# component config attacker impact vector condition
DETAILS_6_2 = (COMPONENT_CONFIG_ATTACKER_PATTERN + IMPACT_BEFORE_STOP_WORDS + IMPACT_PATTERN +
               VECTOR_BEFORE_STOP_WORDS + VECTOR_PATTERN + CONDITION_BEFORE_STOP_WORDS + CONDITION_PATTERN + r'\.$')

# component config attacker impact vector
DETAILS_5_1 = (COMPONENT_CONFIG_ATTACKER_PATTERN + IMPACT_BEFORE_STOP_WORDS + IMPACT_PATTERN +
               VECTOR_BEFORE_STOP_WORDS + VECTOR_PATTERN + r'\.$')

# component config attacker impact root_cause
DETAILS_5_2 = (COMPONENT_CONFIG_ATTACKER_PATTERN + IMPACT_BEFORE_STOP_WORDS + IMPACT_PATTERN +
               ROOT_CAUSE_BEFORE_STOP_WORD + ROOT_CAUSE_PATTERN + r'\.$')


COMPONENT_CONFIG_ATTACKER_PATTERNS = {
    '6_1': DETAILS_6_1,
    '6_2': DETAILS_6_2,
    '5_1': DETAILS_5_1,
    '5_2': DETAILS_5_2
}



import re

from typing import Tuple, Dict
from .common.enums import DetailType
from .patterns import STARTING_DETAILS_MAPPING, DETAILS_PATTERNS_MAPPING
from .patterns.common import REPLACEMENT_PHRASES, NOISE_PATTERNS


class CVEDriller:
    def __init__(self, verbose: bool = False):
        self.verbose = verbose

    def log_info(self, message: str):
        if self.verbose:
            print(message)

    def extract_from_patterns(self, desc: str, patterns: dict):
        for name, pattern in patterns.items():
            match = re.compile(pattern).match(desc)

            if match:
                self.log_info(f"{name} - {pattern}")
                return match.groupdict()

        return None

    @staticmethod
    def normalize_phrases(desc: str) -> str:
        if desc[-1] != '.':
            desc += '.'

        desc = re.sub(r'\s+', ' ', desc)
        desc = re.sub(r"\"", "", desc)

        for old, new in REPLACEMENT_PHRASES.items():
            # TODO: should separate the replacement_phrases into categories and apply only one per each category
            desc = re.sub(old, new, desc)

        return desc

    @staticmethod
    def remove_noise(desc: str) -> str:
        for pattern in NOISE_PATTERNS:
            desc = re.sub(pattern, "", desc)

        return desc

    def __call__(self, desc: str) -> Tuple[DetailType | None, Dict[str, str] | None]:
        desc = desc.strip()
        desc = self.normalize_phrases(desc)
        desc = self.remove_noise(desc)

        last_matching_detail_type = None

        self.log_info(f"Description: {desc}")
        for detail_type, pattern in STARTING_DETAILS_MAPPING.items():
            details_patterns = DETAILS_PATTERNS_MAPPING[detail_type]

            if re.match(pattern, desc):
                last_matching_detail_type = detail_type
                extracted_details = self.extract_from_patterns(desc, details_patterns)
                self.log_info(f"Found starting pattern: {detail_type}")

                if extracted_details:
                    return detail_type, extracted_details

        return last_matching_detail_type, None



from ...common import (CONFIGURATION_PATTERN, CONFIGURATION_BEFORE_STOP_WORDS, IMPACT_BEFORE_STOP_WORDS, IMPACT_PATTERN,
                       ROOT_CAUSE_PATTERN, ROOT_CAUSE_BEFORE_STOP_WORD, WEAKNESS_PATTERN, WEAKNESS_AFTER_STOP_WORDS,
                       CONFIGURATION_AFTER_STOP_WORDS, CONDITION_BEFORE_STOP_WORDS, CONDITION_PATTERN,
                       ATTACKER_BEFORE_STOP_WORDS, ATTACKER_PATTERN, VECTOR_BEFORE_STOP_WORDS, VECTOR_PATTERN)


# weakness configuration root_cause
WEAKNESS_CONFIG_CAUSE_PATTERN = (
        r'(An |A |)' + WEAKNESS_PATTERN + WEAKNESS_AFTER_STOP_WORDS +
        CONFIGURATION_BEFORE_STOP_WORDS + CONFIGURATION_PATTERN + CONFIGURATION_AFTER_STOP_WORDS +
        ROOT_CAUSE_BEFORE_STOP_WORD + ROOT_CAUSE_PATTERN
)

# weakness configuration root_cause condition attacker impact
DETAILS_6_1 = (
        WEAKNESS_CONFIG_CAUSE_PATTERN + CONDITION_BEFORE_STOP_WORDS + CONDITION_PATTERN + ATTACKER_BEFORE_STOP_WORDS +
        ATTACKER_PATTERN + IMPACT_BEFORE_STOP_WORDS + IMPACT_PATTERN + r'\.$'
)

# weakness configuration root_cause attacker impact vector
DETAILS_6_2 = (
        WEAKNESS_CONFIG_CAUSE_PATTERN + ATTACKER_BEFORE_STOP_WORDS + ATTACKER_PATTERN + IMPACT_BEFORE_STOP_WORDS +
        IMPACT_PATTERN + VECTOR_BEFORE_STOP_WORDS + VECTOR_PATTERN + r'\.$'
)

# weakness configuration root_cause impact
DETAILS_4_1 = (
        WEAKNESS_CONFIG_CAUSE_PATTERN + IMPACT_BEFORE_STOP_WORDS + IMPACT_PATTERN + r'\.$'
)


WEAKNESS_CONFIG_CAUSE_PATTERNS = {
    '6_1': DETAILS_6_1,
    '6_2': DETAILS_6_2,
    '4_1': DETAILS_4_1
}



from ...common import (WEAKNESS_PATTERN, WEAKNESS_AFTER_STOP_WORDS, CONFIGURATION_PATTERN, VECTOR_BEFORE_STOP_WORDS,
                       CONFIGURATION_BEFORE_STOP_WORDS, VECTOR_PATTERN, ATTACKER_BEFORE_STOP_WORDS, ATTACKER_PATTERN,
                       IMPACT_BEFORE_STOP_WORDS, IMPACT_PATTERN, COMPONENT_BEFORE_STOP_WORDS, COMPONENT_PATTERN,
                       ROOT_CAUSE_BEFORE_STOP_WORD, ROOT_CAUSE_PATTERN)

# weakness configuration vector
WEAKNESS_CONFIG_VECTOR_PATTERN =  (r'(An |A |)' + WEAKNESS_PATTERN + WEAKNESS_AFTER_STOP_WORDS +
                                   CONFIGURATION_BEFORE_STOP_WORDS + CONFIGURATION_PATTERN + VECTOR_BEFORE_STOP_WORDS +
                                   VECTOR_PATTERN)

# weakness configuration vector attacker impact
DETAILS_5_1 = (WEAKNESS_CONFIG_VECTOR_PATTERN + ATTACKER_BEFORE_STOP_WORDS + ATTACKER_PATTERN +
               IMPACT_BEFORE_STOP_WORDS + IMPACT_PATTERN + r'\.$')

# weakness configuration vector component
DETAILS_4_1 = WEAKNESS_CONFIG_VECTOR_PATTERN + COMPONENT_BEFORE_STOP_WORDS + COMPONENT_PATTERN + r'\.$'

# weakness configuration vector root_cause
DETAILS_4_2 = WEAKNESS_CONFIG_VECTOR_PATTERN + ROOT_CAUSE_BEFORE_STOP_WORD + ROOT_CAUSE_PATTERN + r'\.$'

# weakness configuration vector impact
DETAILS_4_3 = WEAKNESS_CONFIG_VECTOR_PATTERN + IMPACT_BEFORE_STOP_WORDS + IMPACT_PATTERN + r'\.$'

WEAKNESS_CONFIG_VECTOR_PATTERNS = {
    '5_1': DETAILS_5_1,
    '4_1': DETAILS_4_1,
    '4_2': DETAILS_4_2,
    '4_3': DETAILS_4_3
}




from ...common import (CONFIGURATION_PATTERN, ATTACKER_PATTERN, ATTACKER_BEFORE_STOP_WORDS, IMPACT_PATTERN,
                       IMPACT_BEFORE_STOP_WORDS, VECTOR_PATTERN, VECTOR_BEFORE_STOP_WORDS, ROOT_CAUSE_PATTERN,
                       ROOT_CAUSE_BEFORE_STOP_WORD, CONFIGURATION_AFTER_STOP_WORDS, CONDITION_BEFORE_STOP_WORDS,
                       CONDITION_PATTERN)

# configuration attacker impact
CONFIG_ATTACKER_IMPACT_PATTERN = (r'(?:The |)' + CONFIGURATION_PATTERN + CONFIGURATION_AFTER_STOP_WORDS +
                                  ATTACKER_BEFORE_STOP_WORDS + ATTACKER_PATTERN + IMPACT_BEFORE_STOP_WORDS +
                                  IMPACT_PATTERN)

# configuration attacker impact root_cause vector
DETAILS_5_1 = (CONFIG_ATTACKER_IMPACT_PATTERN + ROOT_CAUSE_BEFORE_STOP_WORD + ROOT_CAUSE_PATTERN +
               VECTOR_BEFORE_STOP_WORDS + VECTOR_PATTERN + r'\.$')

# configuration attacker impact vector root_cause
DETAILS_5_2 = (CONFIG_ATTACKER_IMPACT_PATTERN + VECTOR_BEFORE_STOP_WORDS + VECTOR_PATTERN +
               ROOT_CAUSE_BEFORE_STOP_WORD + ROOT_CAUSE_PATTERN + r'\.$')

# configuration attacker impact vector condition
DETAILS_5_3 = (CONFIG_ATTACKER_IMPACT_PATTERN + VECTOR_BEFORE_STOP_WORDS + VECTOR_PATTERN +
               CONDITION_BEFORE_STOP_WORDS + CONDITION_PATTERN + r'\.$')

# configuration attacker impact root_cause
DETAILS_4_1 = CONFIG_ATTACKER_IMPACT_PATTERN + ROOT_CAUSE_BEFORE_STOP_WORD + ROOT_CAUSE_PATTERN + r'\.$'

# configuration attacker impact condition
DETAILS_4_2 = CONFIG_ATTACKER_IMPACT_PATTERN + CONDITION_BEFORE_STOP_WORDS + CONDITION_PATTERN + r'\.$'


CONFIG_ATTACKER_IMPACT_PATTERNS = {
    '5_1': DETAILS_5_1,
    '5_2': DETAILS_5_2,
    '5_3': DETAILS_5_3,
    '4_1': DETAILS_4_1,
    '4_2': DETAILS_4_2
}



import json

from cve_driller.driller import CVEDriller


if __name__ == "__main__":
    driller = CVEDriller(verbose=True)
    cve_id = "CVE-2018-7187"

    input_path = f"data/inputs/CWE-78/comp_config_condition/{cve_id}.txt"
    #description = open(input_path, 'r').read()
    description = "SQL injection vulnerability in members.php in the Members CV (job) module 1.0 for PHP-Fusion, when magic_quotes_gpc is disabled, allows remote authenticated users to execute arbitrary SQL commands via the sortby parameter."
    detail_type, details = driller(description)

    print(details)

    #if details:
    #    output_path = f"data/outputs/{detail_type.value}/{cve_id}.json"

    #    with open(output_path, 'w') as f:
    #        json.dump(details, f, indent=4)
